<?php

namespace LWS\WOOREWARDS\PRO\Events;

// don't call the file directly
if (!defined('ABSPATH')) exit();

/** Earn points when a bought product belong to a specified category. */
class BuyInCategory extends \LWS\WOOREWARDS\Abstracts\Event
implements \LWS\WOOREWARDS\PRO\Events\I_CartPreview
{
	use \LWS\WOOREWARDS\PRO\Events\T_ExcludedProducts;
	use \LWS\WOOREWARDS\PRO\Events\T_Order;
	use \LWS\WOOREWARDS\PRO\Events\T_SponsorshipOrigin;

	function getInformation()
	{
		return array_merge(parent::getInformation(), array(
			'icon'  => 'lws-icon-barcode',
			'short' => __("The customer will earn points for buying products from specific categories.", 'woorewards-pro'),
			'help'  => __("The number of points given is fixed for every product.", 'woorewards-pro'),
		));
	}

	function getDescription($context = 'backend')
	{
		$categories = $this->getProductCategories();
		$msg = '';
		if ($this->getMinProductCount() > 1)
		{
			$msg = _n('Buy at least %2$d products in category %1$s', 'Buy at least %2$d products in categories %1$s', count($categories), 'woorewards-pro');
			return sprintf($msg, $this->getProductCategoriesNames($categories, $context), $this->getMinProductCount());
		}
		else
		{
			$msg = _n("Buy products in category %s", "Buy products in categories %s", count($categories), 'woorewards-pro');
			return sprintf($msg, $this->getProductCategoriesNames($categories, $context));
		}
	}

	protected function getProductCategoriesNames($categories, $context = 'backend', $sep = ', ')
	{
		$descr = '';
		$mid = '';
		foreach ($categories as $cat)
		{
			$descr .= $mid;
			$term = \get_term($cat, 'product_cat');
			if (!$term || is_wp_error($term))
			{
				$descr .= _x("Unknown", "Unknown category", 'woorewards-pro');
			}
			else
			{
				$name = \htmlentities($term->name);

				$url = '';
				if ($context == 'backend')
					$url = \esc_attr(\get_edit_tag_link($cat, 'product_cat'));

				if (empty($url))
				{
					if ($context == 'raw')
						$descr .= $name;
					else
						$descr .= "<b>{$name}</b>";
				}
				else
					$descr .= "<a href='{$url}'>{$name}</a>";
			}
			$mid = $sep;
		}
		return $descr;
	}

	/** add help about how it works */
	function getForm($context = 'editlist')
	{
		$prefix = $this->getDataKeyPrefix();
		$form = parent::getForm($context);

		$label = __("Priority", 'woorewards-pro');
		$tooltip = __("Customer orders will run by ascending priority value.", 'woorewards-pro');
		$str = <<<EOT
		<div class='field-help'>$tooltip</div>
		<div class='lws-$context-opt-title label'>$label<div class='bt-field-help'>?</div></div>
		<div class='lws-$context-opt-input value'>
			<input type='text' id='{$prefix}event_priority' name='{$prefix}event_priority' placeholder='10' size='5' />
		</div>
EOT;
		$phb0 = $this->getFieldsetPlaceholder(false, 0);
		$form = str_replace($phb0, $str . $phb0, $form);

		$form .= $this->getFieldsetBegin(2, __("Options", 'woorewards-pro'));
		// multiply by quantity
		$label   = _x("Quantity Multiplier", "Buy In Category Event", 'woorewards-pro');
		if ($context == 'achievements')
		{
			$tooltip = __("If checked, action will be counted once per product in the cart meeting the conditions. Otherwise, only once per order.", 'woorewards-pro');
		}
		else
		{
			$tooltip = __("If checked, points will be earned for each product in the cart meeting the conditions. Otherwise, points will be earned only once per order", 'woorewards-pro');
		}
		$checked = $this->isQtyMultiply() ? 'checked' : '';
		$form .= "<div class='field-help'>$tooltip</div>";
		$form .= "<div class='lws-$context-opt-title label'>$label<div class='bt-field-help'>?</div></div>";
		$form .= "<div class='lws-$context-opt-input value'><input class='lws_checkbox' type='checkbox' id='{$prefix}qty_multiply' name='{$prefix}qty_multiply' $checked/></div>";

		// value
		$label = _x("Minimum product count", "Coupon Unlockable", 'woorewards-pro');
		$value = \esc_attr($this->getMinProductCount());
		$form .= "<div class='lws-$context-opt-title label'>$label</div>";
		$form .= "<div class='lws-$context-opt-input value'><input type='text' id='{$prefix}product_count' name='{$prefix}product_count' value='$value' placeholder='1' pattern='\\d*(\\.|,)?\\d*' /></div>";

		$form .= $this->getFieldsetEnd(2);

		$form .= $this->getFieldsetBegin(3, __("Product categories", 'woorewards-pro'), 'span2');

		// The product category
		$label   = _x("Categories", "Buy In Category Event", 'woorewards-pro');
		$form .= "<div class='lws-$context-opt-title label'>$label</div>";
		$form .= "<div class='lws-$context-opt-input value'>";
		$form .= \LWS\Adminpanel\Pages\Field\LacChecklist::compose($prefix . 'product_cat', array(
			'comprehensive' => true,
			'predefined' => 'taxonomy',
			'spec' => array('taxonomy' => 'product_cat'),
			'value' => $this->getProductCategories()
		));
		$form .= "</div>";
		$form .= $this->getFieldsetEnd(3);
		$form =  $this->filterForm($form, $prefix, $context, 2);
		return $this->filterSponsorshipForm($form, $prefix, $context, 10);
	}

	function getData()
	{
		$prefix = $this->getDataKeyPrefix();
		$data = parent::getData();
		$data[$prefix . 'product_cat'] = base64_encode(json_encode($this->getProductCategories()));
		$data[$prefix . 'product_count'] = $this->getMinProductCount();
		$data[$prefix . 'qty_multiply'] = $this->isQtyMultiply() ? 'on' : '';
		$data[$prefix . 'event_priority'] = $this->getEventPriority();
		$data = $this->filterSponsorshipData($data, $prefix);
		return $this->filterData($data, $prefix);
	}

	function submit($form = array(), $source = 'editlist')
	{
		$prefix = $this->getDataKeyPrefix();
		$values = \apply_filters('lws_adminpanel_arg_parse', array(
			'post'     => ($source == 'post'),
			'values'   => $form,
			'format'   => array(
				$prefix . 'product_cat' => array('D'),
				$prefix . 'product_count' => 'D',
				$prefix . 'qty_multiply' => 's',
				$prefix . 'event_priority'   => 'd',
			),
			'defaults' => array(
				$prefix . 'product_count' => '1',
				$prefix . 'qty_multiply' => '',
				$prefix . 'event_priority'   => $this->getEventPriority(),
			),
			'required' => array(
				$prefix . 'product_cat' => true
			),
			'labels'   => array(
				$prefix . 'product_cat'   => __("Product category", 'woorewards-pro'),
				$prefix . 'product_count' => __("Minimum product count", 'woorewards-pro'),
				$prefix . 'qty_multiply' => __("Quantity Multiplier", 'woorewards-pro'),
				$prefix . 'event_priority'   => __("Event Priority", 'woorewards-pro'),
			)
		));
		if (!(isset($values['valid']) && $values['valid']))
			return isset($values['error']) ? $values['error'] : false;

		$valid = parent::submit($form, $source);
		if ($valid === true)
			$valid = $this->optSponsorshipSubmit($prefix, $form, $source);
		if ($valid === true && ($valid = $this->optSubmit($prefix, $form, $source)) === true)
		{
			$this->setProductCategories($values['values'][$prefix . 'product_cat']);
			$this->setMinProductCount($values['values'][$prefix . 'product_count']);
			$this->setQtyMultiply($values['values'][$prefix . 'qty_multiply']);
			$this->setEventPriority($values['values'][$prefix . 'event_priority']);
		}

		return $valid;
	}

	function isQtyMultiply()
	{
		return isset($this->qtyMultiply) && $this->qtyMultiply;
	}

	public function setQtyMultiply($yes = false)
	{
		$this->qtyMultiply = boolval($yes);
		return $this;
	}

	function getMinProductCount()
	{
		return isset($this->minProductCount) ? $this->minProductCount : 1;
	}

	function setMinProductCount($n)
	{
		$this->minProductCount = max(1, intval($n));
		return $this;
	}

	function getProductCategories()
	{
		return isset($this->productCategories) ? $this->productCategories : array();
	}

	/** @param $categories (array|string) as string, it should be a json base64 encoded array. */
	function setProductCategories($categories = array())
	{
		if (!is_array($categories))
			$categories = @json_decode(@base64_decode($categories));
		if (is_array($categories))
			$this->productCategories = $categories;
		return $this;
	}

	protected function _fromPost(\WP_Post $post)
	{
		$this->setProductCategories(\get_post_meta($post->ID, 'wre_event_product_cat', true));
		$this->setMinProductCount(\get_post_meta($post->ID, 'wre_event_min_product_count', true));
		$this->setQtyMultiply(\get_post_meta($post->ID, 'wre_event_qty_multiply', true));
		$this->setEventPriority($this->getSinglePostMeta($post->ID, 'wre_event_priority', $this->getEventPriority()));
		$this->optSponsorshipFromPost($post);
		$this->optFromPost($post);
		return $this;
	}

	protected function _save($id)
	{
		\update_post_meta($id, 'wre_event_product_cat', $this->getProductCategories());
		\update_post_meta($id, 'wre_event_min_product_count', $this->getMinProductCount());
		\update_post_meta($id, 'wre_event_qty_multiply', $this->isQtyMultiply() ? 'on' : '');
		\update_post_meta($id, 'wre_event_priority', $this->getEventPriority());
		$this->optSponsorshipSave($id);
		$this->optSave($id);
		return $this;
	}

	/** @return a human readable type for UI */
	public function getDisplayType()
	{
		return _x("Buy in category", "getDisplayType", 'woorewards-pro');
	}

	function getEventPriority()
	{
		return isset($this->eventPriority) ? \intval($this->eventPriority) : 50;
	}

	public function setEventPriority($priority)
	{
		$this->eventPriority = \intval($priority);
		return $this;
	}

	/** Add hook to grab events and add points. */
	protected function _install()
	{
		\add_filter('lws_woorewards_wc_order_done_' . $this->getPoolName(), array($this, 'orderDone'), $this->getEventPriority());
	}

	function orderDone($order)
	{
		$sponsorship = new \LWS\WOOREWARDS\PRO\Core\Sponsorship();
		$this->sponsorship = $sponsorship->getUsersFromOrder($order->order, false);

		if (!$this->acceptOrder($order->order))
			return $order;
		if (!$this->isValidOrigin($this->sponsorship->origin))
			return $order;
		if (empty($categories = $this->getProductCategories()))
			return $order;
		if (!($userId = $order->order->get_customer_id('edit')))
		{
			$user = \get_user_by('email', $order->order->get_billing_email());
			$userId = ($user && $user->ID) ? $user->ID : 0;
		}
		if (!$userId)
			return $order;

		$noEarning = $this->getExclusionFromOrder($order->order);
		$floor = $this->getMinProductCount();
		$boughtInCat = 0;
		foreach ($order->items as $item)
		{
			$product = \LWS\WOOREWARDS\PRO\Conveniences::instance()->getProductFromOrderItem($order->order, $item->item);
			if ($product && $this->isProductInCategory($product, $categories))
			{
				$qty = $item->item->get_quantity();
				$qty = $this->useExclusion($noEarning, $product, $qty);
				$boughtInCat += $qty;
			}
		}
		if ($boughtInCat < $floor)
			return $order;

		//Take quantity ordered into account or not
		if ($pointsCount = \apply_filters('trigger_' . $this->getType(), $this->isQtyMultiply() ? $boughtInCat : 1, $this, $order->order))
		{
			$reason = \LWS\WOOREWARDS\Core\Trace::byOrder($order->order)
				->setProvider($order->order->get_customer_id('edit'))
				->setReason(array(
					(1 == count($categories) ? "Product bought in category %s" : "Product bought in categories %s"),
					$this->getProductCategoriesNames($categories, 'raw')
				), 'woorewards-pro');
			$this->addPoint($userId, $reason, $pointsCount);
		}
		return $order;
	}

	/** Never call, only to have poedit/wpml able to extract the sentance. */
	private function poeditDeclare()
	{
		__("Product bought in category %s", 'woorewards-pro');
		__("Product bought in categories %s", 'woorewards-pro');
	}

	private function isProductInCategory($product, $whiteList, $taxonomy = 'product_cat')
	{
		$product_cats = \wc_get_product_cat_ids($product->get_id());
		if (!empty($parentId = $product->get_parent_id()))
			$product_cats = array_merge($product_cats, \wc_get_product_cat_ids($parentId));

		// If we find an item with a cat in our allowed cat list, the product is valid.
		return !empty(array_intersect($product_cats, $whiteList));
	}

	/**	Event categories, used to filter out events from pool options.
	 *	@return array with category_id => category_label. */
	public function getCategories()
	{
		return array_merge(parent::getCategories(), array(
			'woocommerce' => __("WooCommerce", 'woorewards-pro'),
			'product'  => __("Product", 'woorewards-pro')
		));
	}

	function getPointsForProduct(\WC_Product $product)
	{
		if (!$this->acceptOrigin('checkout'))
			return 0;

		if ($this->getMinProductCount() == 1 && !empty($categories = $this->getProductCategories()))
		{
			if ($this->isProductInCategory($product, $categories))
				return $this->getMultiplier();
		}
		return 0;
	}

	/** Multiplier is registered by Pool, it is applied to the points generated by the event. */
	public function getMultiplier($context = 'edit')
	{
		$mul = parent::getMultiplier($context);
		if ($context == 'view' && $mul > 0 && $this->isQtyMultiply())
		{
			$mul = sprintf(
				_x("%s / item", "Points per item", 'woorewards-pro'),
				\LWS_WooRewards::formatPointsWithSymbol($mul, $this->getPoolName())
			);
		}
		return $mul;
	}

	function getPointsForCart(\WC_Cart $cart)
	{
		$sponsorship = new \LWS\WOOREWARDS\PRO\Core\Sponsorship();
		$this->sponsorship = $sponsorship->getCurrentReferral();

		if (!$this->acceptOrigin('checkout'))
			return 0;
		if (!$this->isValidOrigin($this->sponsorship->origin))
			return 0;

		if (!empty($categories = $this->getProductCategories()))
		{
			$noEarning = $this->getExclusionFromCart($cart);
			$floor = $this->getMinProductCount();
			$boughtInCat = 0;
			foreach ($cart->get_cart() as $item)
			{
				if (isset($item['product_id']) && !empty($product = \wc_get_product($item['product_id'])))
				{
					if ($this->isProductInCategory($product, $categories))
					{
						$qty = isset($item['quantity']) ? intval($item['quantity']) : 1;
						$qty = $this->useExclusion($noEarning, $product, $qty);
						$boughtInCat += $qty;
					}
				}
			}
			if ($boughtInCat >= $floor)
			{
				$pointsCount = ($this->isQtyMultiply() ? $boughtInCat : 1);
				return $this->getMultiplier() * $pointsCount;
			}
		}
		return 0;
	}

	function getPointsForOrder(\WC_Order $order)
	{
		$sponsorship = new \LWS\WOOREWARDS\PRO\Core\Sponsorship();
		$this->sponsorship = $sponsorship->getUsersFromOrder($order, false);

		if (!$this->acceptOrder($order))
			return 0;
		if (!$this->isValidOrigin($this->sponsorship->origin))
			return 0;

		if (!empty($categories = $this->getProductCategories()))
		{
			$noEarning = $this->getExclusionFromOrder($order);
			$floor = $this->getMinProductCount();
			$boughtInCat = 0;
			foreach ($order->get_items() as $item)
			{
				if (isset($item['product_id']) && !empty($product = \wc_get_product($item['product_id'])))
				{
					if ($this->isProductInCategory($product, $categories))
					{
						$qty = isset($item['quantity']) ? intval($item['quantity']) : 1;
						$qty = $this->useExclusion($noEarning, $product, $qty);
						$boughtInCat += $qty;
					}
				}
			}
			if ($boughtInCat >= $floor)
			{
				$pointsCount = ($this->isQtyMultiply() ? $boughtInCat : 1);
				return $this->getMultiplier() * $pointsCount;
			}
		}
		return 0;
	}
}
